import java.applet.Applet;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.lang.Thread;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Vector;

/**
 * A Control applet for Infra-Red Remote Controller Interface.
 * @version 	1.0,  9 Jul 1997
 * @author Jan Kratochvil
 */

public
class Control extends Applet {
int IRCOND_PORT=3250;
Dimension CurDim,OrigDim=null;
Image MainImage=null;
boolean Active=false;

class eRectangle extends Rectangle {
eRectangle(Rectangle r,int o) {
	setBounds(r.x+o,r.y+o,r.width-2*o,r.height-2*o);
}
	}

void repaint(Rectangle r)
	{ repaint(r.x,r.y,r.width,r.height); }

void drawBox(Graphics gfx,eRectangle rs,eRectangle rf) {
int d=rf.x-rs.x;
	gfx.fillRect(rs.x,rs.y          ,rs.width,d); //upper
	gfx.fillRect(rs.x,rf.y+rf.height,rs.width,d); //lower
	gfx.fillRect(rs.x         ,rf.y,d,rf.height); //left
	gfx.fillRect(rf.x+rf.width,rf.y,d,rf.height); //right
}

class IRConThread extends Thread {
Vector cmds=new Vector();

public void run() {
Socket s;
String host;
InputStreamReader fr;
OutputStreamWriter fw;
char c1,c2=' ';
try {
	if ((host=getParameter("ircond"))==null) host=getDocumentBase().getHost();
	showStatus("Control: Connecting to ircond at "+host+"...");
	s=new Socket(InetAddress.getByName(host),IRCOND_PORT);
	fr=new InputStreamReader(s.getInputStream());
	fw=new OutputStreamWriter(s.getOutputStream());
	for (;;) {
		do { c1=c2; c2=(char)fr.read(); } while (c1!='\n'||c2!='>');
		if (cmds.isEmpty()) {
			updateAreaStatus();
			if (Active) {
				Active=false;
				if (AreaMove!=null&&Timer!=null) repaint(AreaMove.rectActive());
				}
			synchronized(this) { wait(60000); } // One minute for timeout
			if (cmds.isEmpty()) {
				IRCon=null;
				showStatus("Control: Closing connection to "+host+" due to user inactivity.");
				s.close();
				stop();
				}
			}
		showStatus("Control: Sending command "+cmds.firstElement()+" to ircond at "+host+"...");
		fw.write(cmds.firstElement()+"\n");
		fw.flush();
		cmds.removeElementAt(0);
		}
	} catch (Exception e) { showStatus(e.toString()); return; }
}
synchronized void send(String cmd) {
	cmds.addElement(cmd);
	notify();
}
	}
IRConThread IRCon=null;

class Area extends Rectangle {
	String action;
	Rectangle osiz;
void UpdateSize() {
	setBounds(osiz.x    *CurDim.width/OrigDim.width,osiz.y     *CurDim.height/OrigDim.height,
	          osiz.width*CurDim.width/OrigDim.width,osiz.height*CurDim.height/OrigDim.height);
}
Area(int [] ndim,String nact) {
	osiz=new Rectangle(ndim[0],ndim[1],ndim[2]+1-ndim[0],ndim[3]+1-ndim[1]);
	action=nact;
}
eRectangle rectMove()
	{ return new eRectangle(this,-2); }
eRectangle rectDown()
	{ return new eRectangle(this,0); }
eRectangle rectActive()
	{ return new eRectangle(this,+2); }
void drawMove(Graphics gfx) {
	gfx.setColor(new Color(0,255,0));
	drawBox(gfx,rectMove(),rectDown());
}
void drawDown(Graphics gfx) {
	gfx.setColor(new Color(255,255,0));
	drawBox(gfx,rectDown(),rectActive());
}
void drawActive(Graphics gfx) {
eRectangle r=rectActive();
	gfx.setColor(new Color(255,0,0));
	gfx.fillRect(r.x,r.y,r.width,r.height);
}
void activate() {
	if (!Active) {
		Active=true;
		repaint(rectActive());
		}
	if (IRCon==null) {
		IRCon=new IRConThread();
		IRCon.setDaemon(true);
		IRCon.start();
		}
	IRCon.send(action);
}
	}

Area AreaMove=null;
Area Arg_Area[];
int AreaLength=0;

void updateAreaStatus() {
	if (AreaMove!=null) showStatus(AreaMove.action);
	else showStatus(null);
}

class InvalidParametersException extends Exception {
public InvalidParametersException()
	{ super("Invalid parameters!"); }
public InvalidParametersException(int num,String s)
	{ super("Parameter "+(num+1)+" is not valid"+(s==null?"":" ("+s+")")+"!"); }
	}

class TimeThread extends Thread {
	long delay;
TimeThread(long ndelay) {
	delay=ndelay;
}
public void run() {
	for (;;) {
		try { sleep(delay); }
		catch (InterruptedException e) { System.out.println("Control: Interrupted at TimeThread - shouldn't happen!"); }
		AreaMove.activate();
		}
}
	}
TimeThread Timer=null;

Area whichArea(Point pnt) {
int i;
	for (i=0;i<AreaLength;i++)
		if (Arg_Area[i].contains(pnt)) return Arg_Area[i];
	return null;
}

void geninit() throws IOException,InvalidParametersException {
int i;
	MainImage=getImage(getCodeBase(),getParameter("ImageName"));
	Arg_Area=new Area[Integer.parseInt(getParameter("NumAreas"))];
	for (i=0;i<Arg_Area.length;i++) {
int token=0,tokens[]=new int[4];
StreamTokenizer st=new StreamTokenizer(new StringReader(getParameter("Area"+i)));
String action=null;
		try { while (st.nextToken()!=st.TT_EOF) {
			switch (st.ttype) {
			case st.TT_NUMBER:
				if (token<4) tokens[token]=new Double(st.nval).intValue();
				else throw new InvalidParametersException(token,"number required");
				break;
			case st.TT_WORD:
				if (token==4) action=st.sval;
				else throw new InvalidParametersException(token,"action string required");
				break;
			default:
				throw new InvalidParametersException(token,st.ttype+" found");
			}
			token++;
			}
		} catch (IOException e) { throw e; }
		if (token!=5) throw new InvalidParametersException();
		Arg_Area[i]=new Area(tokens,action);
		AreaLength++;
		}
}

public void init() { /* Why the hell can't init() throw?!? */
	try { geninit(); }
	catch (Exception e) { showStatus(e.toString()); }
}

public void paint(Graphics gfx) {
int i;
  CurDim = size();
	if (OrigDim==null) OrigDim=CurDim;
	for (i=0;i<AreaLength;i++)
		Arg_Area[i].UpdateSize();
	gfx.drawImage(MainImage,0,0,CurDim.width,CurDim.height,this);
	if (AreaMove==null) return;
	AreaMove.drawMove(gfx);
	if (Timer==null) return;
	AreaMove.drawDown(gfx);
	if (!Active) return;
	AreaMove.drawActive(gfx);
}
public void update(Graphics gfx)
	{ paint(gfx); } // Will someone give it at least a call?

void userUpdate(Point pt,boolean mouse) {
Area area;
	if ((area=whichArea(pt))!=AreaMove) {
		if (AreaMove!=null) repaint(AreaMove.rectMove());
		AreaMove=area;
		if (AreaMove!=null) repaint(AreaMove.rectMove());
		updateAreaStatus();
		}
	if ((mouse&&AreaMove!=null)==(Timer!=null)) return;
	if (AreaMove!=null) repaint(AreaMove.rectDown());
	if (Timer==null) {
		AreaMove.activate();
		Timer=new TimeThread(500);
		Timer.setDaemon(true);
		Timer.start();
		}
	else { Timer.stop(); Timer=null; }
}

public boolean mouseMove(Event ev,int x,int y) {
	userUpdate(new Point(x,y),false);
	return true;
}

public boolean mouseDrag(Event ev, int x, int y) {
	userUpdate(new Point(x,y),true);
	return true;
}

public boolean mouseUp(Event ev, int x, int y) {
	userUpdate(new Point(x,y),false);
	return true;
}

public boolean mouseDown(Event ev, int x, int y) {
	userUpdate(new Point(x,y),true);
	return true;
}

public boolean mouseExit(Event ev, int x, int y) {
	userUpdate(new Point(-1,-1),false);
	return true;
}

public boolean mouseEnter(Event ev, int x, int y) {
	userUpdate(new Point(x,y),false);
	return true;
}

/* {got,lost}Focus somehow required... WHY?!? */
public boolean gotFocus(Event ev,Object what)
	{ return true; }
public boolean lostFocus(Event ev,Object what)
	{ return true; }

public String getAppletInfo() {
	return "Infra-Red Controller Control by Jan Kratochvil";
}

public String [][] getParameterInfo() {
String info[][]={
{"ImageName","String"                ,"Main image"                           },
{"NumAreas" ,"int"                   ,"Number of supplied AreaX-es"          },
{"AreaX"    ,"int int int int String","Left Top Right Bottom Command-to-send"},
{"ircond"   ,"String"                ,"Hostname of ircond daemon to contact" },
};
	return info;
}

}
