#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "util.h"
#include "ircode-db.h"

char protbuf[4096];

enum { DEC_AM, DEC_PULS, DEC_DIG, DEC_PROT } decode;

static double fromstr(const char *arg)
{
char *s;
double val;
	val=strtod(arg,&s);
	if (*s!='\0') {
		fprintf(stderr,"Number convert error: \"%s\"\n",arg);
		exit(EXIT_FAILURE);
		}
	return val;
}

static inline void outlong(unsigned long val)
{
	if (fwrite(&val,sizeof(val),1,stdout)!=1) {
		fprintf(stderr,"Write error (%d=%s)\n",errno,strerror(errno));
		exit(EXIT_FAILURE);
		}
}

static int resolve(const char *vendor,const char *bits,int flg)
{
const char *cs;
	if ((cs=irdb_getname(vendor,bits)))
		{ puts(cs); return 1; }
	else if (flg) return 0;
	else printf("%-36s=%s\n",vendor,bits);
	return 1;
}

static int tzoz_parse(const char *vendor,int bits1,int bits0,int flg)
{
static const char dig0[]="01",dig1[]="0001";
const char *cs;
char decbuf[sizeof(protbuf)],*decptr=decbuf;
	for (cs=protbuf;*cs=='1'&&bits1;cs++,bits1--);
	if (bits1) return 0;
	for (;*cs=='0'&&bits0;cs++,bits0--);
	if (bits0) return 0;
	while (*cs&&decptr+1<decbuf+sizeof(decbuf))
		     if (!memcmp(dig0,cs,sizeof(dig0)-1)) { cs+=sizeof(dig0)-1; *decptr++='0'; }
		else if (!memcmp(dig1,cs,sizeof(dig1)-1)) { cs+=sizeof(dig1)-1; *decptr++='1'; }
		else return 0;
	if (*cs) return 0;
	*decptr='\0';
	return flg==2?(int)strdup(decbuf):resolve(vendor,decbuf,flg);
}

static inline int decode_panasonic(void)
{
	return tzoz_parse("Panasonic",8,1,0);
}

static inline int decode_aiwa(void)
{
	return tzoz_parse("AIWA",17,5,0);
}

static inline int decode_maspro_toshiba(void)
{
char *buf=(char *)tzoz_parse(NULL,16,5,2),*s;
	if (!buf) return 0;
	if (buf[0]!='1'||strlen(buf)!=33) { free(buf); return 0; }
	for (s=buf+ 9;s<buf+17;s++) *s=*s=='0'?'1':'0';
	for (s=buf+25;s<buf+33;s++) *s=*s=='0'?'1':'0';
	if (memcmp(buf+1,buf+9,8)||memcmp(buf+17,buf+25,8)) { free(buf); return 0; }
	memmove(buf,buf+1,8); memmove(buf+8,buf+17,8); buf[16]='\0';
	if (!resolve("Maspro" ,buf,1))
	if (!resolve("Toshiba",buf,1))
		resolve("Maspro/Toshiba",buf,0);
	free(buf);
	return 1;
}

static inline int puls_parse(const char *vendor,int frag,int flg)
{
const char *cs,*cs2;
char decbuf[frag+1];
	if (strlen(protbuf)%frag) return 0;
	for (cs=protbuf,cs2=protbuf+frag;*cs2;cs++,cs2++)
		if (*cs!=*cs2) return 0;
	memcpy(decbuf,protbuf,frag); decbuf[frag]='\0';
	return flg==2?(int)strdup(decbuf):resolve(vendor,decbuf,flg);
}

static inline int decode_korting(void)
{
	return puls_parse("Korting",7,0);
}

static inline int decode_sharp(void)
{
char *buf=(char *)puls_parse(NULL,30,2),*s;
int ret;
	if (!buf) return 0;
	for (s=buf+20;*s;s++) *s=*s=='0'?'1':'0';
	if (!memcmp(buf,buf+15,15)&&!memcmp(buf,"11001",5))
		ret=resolve("Sharp",buf+20,0);
	else ret=0;
	free(buf);
	return ret;
}

static inline void protdecode(int argc,char **argv)
{
char *s;
int val=0;
	if (!fgets(protbuf,sizeof(protbuf),stdin)||!(s=strchr(protbuf,'\n'))||*s!='\n') {
		fprintf(stderr,"Error reading digit-stream!\n");
		exit(EXIT_FAILURE);
		}
	*s='\0';
	if (!*protbuf) {
		fprintf(stderr,"Empty digit-stream!\n");
		exit(EXIT_FAILURE);
		}
	for (s=protbuf;*s;s++) {
		if (*s!='0'&&*s!='1') {
			fprintf(stderr,"Non-digit character `%c' in digit buffer at %u!\n",*s,s-protbuf);
			exit(EXIT_FAILURE);
			}
		}
	irdb_load(NULL);
	val|=decode_panasonic();
	val|=decode_korting();
	val|=decode_aiwa();
	val|=decode_maspro_toshiba();
	val|=decode_sharp();
	if (!val) {
		fprintf(stderr,"Unrecognized digit stream: \"%s\"\n",protbuf);
		exit(EXIT_FAILURE);
		}
	exit(EXIT_SUCCESS);
}

int main(int argc,char **argv)
{
unsigned long val[2],valsum,cumul,ioffs,ooffs;
long ledg[2],hedg[2],pedg[2];
double plow,phigh,ppuls=0,ptol,dval;
#define RANGED(x,edg) ((x)>=edg[0]&&(x)<=edg[1])
	if (argc==1) {
		fprintf(stderr,"Usage: decode {{AM|DIG} <low delay> <high delay> <tolerance%%> | PULS <low for 0> <low for 1> <high> <tolerance %%> | PROT}\n");
		exit(EXIT_FAILURE);
		}
	     if (!strcasecmp(argv[1],"am"  )) decode=DEC_AM;
	else if (!strcasecmp(argv[1],"puls")) decode=DEC_PULS;
	else if (!strcasecmp(argv[1],"dig" )) decode=DEC_DIG;
	else if (!strcasecmp(argv[1],"prot")) decode=DEC_PROT;
	else {
		fprintf(stderr,"Decoding-type should be AM, DIG or PROT!\n");
		exit(EXIT_FAILURE);
		}
	if (decode==DEC_PROT) protdecode(argc,argv);
	if ((decode==DEC_PULS&&argc!=6)||(decode!=DEC_PULS&&argc!=5)) {
		fprintf(stderr,"Inappropriate number of parameters - see help!\n");
		exit(EXIT_FAILURE);
		}
	plow =fromstr(argv[2]);
	if (plow<0) {
		fprintf(stderr,"Low delay value must be non-negative!\n");
		exit(EXIT_FAILURE);
		}
	phigh=fromstr(argv[3]);
	if (phigh<0) {
		fprintf(stderr,"High delay value must be non-negative\n");
		exit(EXIT_FAILURE);
		}
	if (decode==DEC_PULS) {
		ppuls=fromstr(argv[4]);
		if (ppuls<0) {
			fprintf(stderr,"Pulse value must be non-negative\n");
			exit(EXIT_FAILURE);
			}
		}
	ptol =fromstr(decode==DEC_PULS?argv[5]:argv[4]);
	if (ptol<0||ptol>=100) {
		fprintf(stderr,"Tolerance must be in range <0..100) !\n");
		exit(EXIT_FAILURE);
		}
	ledg[0]=plow *(100-ptol)/100;
	ledg[1]=plow *(100+ptol)/100;
	hedg[0]=phigh*(100-ptol)/100;
	hedg[1]=phigh*(100+ptol)/100;
	if (decode==DEC_PULS) {
		pedg[0]=ppuls*(100-ptol)/100;
		pedg[1]=ppuls*(100+ptol)/100;
		}
	fprintf(stderr,"%s decoder, cutting-edges: low <%ld..%ld>, high <%ld..%ld>\n",
			(decode==DEC_AM?"Amplitude-modulation":(decode==DEC_DIG?"Amplitude-switching digits":"Korting style short pulses")),ledg[0],ledg[1],hedg[0],hedg[1]);
	if (ledg[0]<=0)
		fprintf(stderr,"Warning: Low edge for low delay <=0 !\n");
	if (hedg[0]<=0)
		fprintf(stderr,"Warning: Low edge for high delay <=0 !\n");
	if (decode==DEC_PULS&&pedg[0]<=0)
		fprintf(stderr,"Warning: Low edge for pulse <=0 !\n");
	ioffs=0; ooffs=0;
	cumul=0;
	while (fread(val,sizeof(val),1,stdin)==1) {
		if (!ioffs) val[0]=0;
		switch (decode) {
int i;
		case DEC_AM:
			if (!ioffs) outlong(0);
			valsum=val[0]+val[1];
			if (val[0]<ledg[0]) {
				fprintf(stderr,"Offset 0x%04lx->0x%04lx: Low delay (%lu) lower than low delay low edge, taking as short!\n",ioffs,ooffs,val[0]);
				val[0]=ledg[0];
				}
			if (!RANGED(val[1],hedg))
				fprintf(stderr,"Offset 0x%04lx->0x%04lx: High delay (%lu) not in range! HUH? Messed up hardware?!? Stucked in active!!!\n",ioffs,ooffs,val[1]);
			if (val[0]>ledg[1]) {
				outlong(cumul); /* High value */
				cumul=0;
				outlong(val[0]); /* Low value */
				ooffs+=sizeof(val);
				}
			else cumul+=valsum;
			break;
		case DEC_DIG:
			for (i=0;i<2;i++) {
				dval=val[i]/(i?phigh:plow);
				if (fabs(dval-rint(dval))>ptol/100)
					fprintf(stderr,"Offset 0x%04lx->%lu: %s delay (%lu) not in range (dval=%g)!\n",ioffs,ooffs,(i?"High":"Low"),val[i],dval);
				ooffs+=cumul=rint(dval);
				while (cumul--)
					putchar('0'+i);
				}
			break;
		case DEC_PULS:
			if (!RANGED(val[1],pedg))
				fprintf(stderr,"Offset 0x%04lx->0x%04lx: Pulse width (%lu) not in range! HUH? Messed up hardware?!? Stucked in active!!!\n",ioffs,ooffs,val[1]);
			if (ioffs&&!RANGED(val[0],ledg)&&!RANGED(val[0],hedg))
				fprintf(stderr,"Offset 0x%04lx->0x%04lx: Delay width (%lu) not in any range! Maybe a central delay, ignoring!!!\n",ioffs,ooffs,val[0]);
			else if (ioffs)
				putchar('0'+RANGED(val[0],hedg));
			break;
		default:
			}
		ioffs+=sizeof(val);
		}
	if (decode==DEC_AM) outlong(cumul);
	else putchar('\n');
	return(EXIT_SUCCESS);
}
