/*
 *	Infra-Red remote controller interface for Linux	
 *
 *	Copyright (C) 1997 Jan Kratochvil
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Based on other minimal char device drivers, like Paul Gortmaker's
 *	rtc, Alan's watchdog, Ted's random, etc. etc.
 *
 *  You probably will need to use 600 mode on this device as
 *  reading/writing is very system unfriendly. User arbitration should
 *  be solved in higher level (such as ircond).
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/lp.h>
#include <linux/kernel.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/poll.h>

#include "ircon.h"

#define IRC_VERSION		"0.99"

#define FIXUSEC(x) (((x)*HZ+500000)/1000000)
#define BACKUSEC(x) (((x)*1000000+HZ/2)/HZ)

#define IRC_IO_EXTENT	3
/* Space length to end reading sequence (in usec/HZ). */
#define DEFAULT_READ_TIMEOUT 500000
/* I use IRCon on dumb Hercules/MDA parallel port. */
#define DEFAULT_BASE_IO 0x3BC

/* Output signal generation with disabled interrupts.
 * Unfortunately this is MANDATORY to work as expected. */
#define USE_CLI

/*
 *	We sponge a minor off of the misc major. No need slurping
 *	up another valuable major dev number for this.
 */

/* misc-minor # until it gets officially registered. */
#ifndef IRCON_MINOR
#define IRCON_MINOR 174
#endif

#define udelayp(usecs) __udelay((usecs),__udelay_val/IRCON_VAL_PRECISION)

int base_io=DEFAULT_BASE_IO;

#define IRC_PORT(x) (base_io+(x))
#undef  r_dtr
#define r_dtr() inb(IRC_PORT(0))
#undef  r_str
#define r_str() inb(IRC_PORT(1))
/*#define r_str() (*(unsigned char *)irc_read)*/
#undef  r_ctr
#define r_ctr() inb(IRC_PORT(2))
#undef  w_dtr
#define w_dtr(x) outb((x),IRC_PORT(0))
/*      w_str(x) is not allowed! */
#undef  w_ctr
#define w_ctr(x) outb((x),IRC_PORT(2))

#define PAIRSIZE (2*sizeof(unsigned long))

static unsigned long irc_read_timeout=FIXUSEC(DEFAULT_READ_TIMEOUT),
										 irc_calibration;

static enum { ILEV_NONE, ILEV_MISCREG }
	init_level=ILEV_NONE;
static unsigned char old_ppctr;

static long long irc_llseek(struct inode *inode, struct file *file,
			    loff_t offset, int origin);

static long irc_read(struct inode *inode, struct file *file,
			char *buf, unsigned long count);

static long irc_write(struct inode *inode, struct file *file,
			const char *buf, unsigned long count);

static int irc_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg);

/*
 *	Bits in irc_status. - currently unused
 */

/*unsigned char irc_status = 0;		* bitmapped status byte.	*/
long custom_timeout;

/* wait_chg() shouldn't be inlined - needed for calibration accuracy! */
/* -finline-functions wouldn't optimize anything anyway in this driver. */
static unsigned long wait_chg_ret;
static int wait_chg_sig;
static unsigned long wait_chg(unsigned char end_val)
{
unsigned long timeout=jiffies+(custom_timeout?custom_timeout:irc_read_timeout);
	custom_timeout=0;
	wait_chg_ret=1; wait_chg_sig=0;
	while ((r_str()&LP_PBUSY)!=end_val) {
		if (jiffies>=timeout) return 0;
		wait_chg_ret++;
		if (current->signal&~current->blocked) {
			wait_chg_sig=1;
			return 0;
			}
		}
	return wait_chg_ret;
}

/*
 *	Now all the various file operations that we export.
 */

static long long irc_llseek(struct inode *inode, struct file *file,
			    loff_t offset, int origin)
{
	return -ESPIPE;
}

static long irc_read(struct inode *inode, struct file *file, char *buf,
		     unsigned long count)
{
unsigned long *p=(void *)buf,total=0,val;
	count/=PAIRSIZE;
	custom_timeout=-1; /* only break-out helps us */
	while (count--) {
		if (!(val=wait_chg(0))) break;
		if (put_user(val,p++)) return -EFAULT;
		val=wait_chg(LP_PBUSY);
		if (wait_chg_sig) { p--; break; }
		total++;
		if (put_user(val,p++)) return -EFAULT;
		if (!val) break;
		}
	if (!total) return -EINTR;
	for (count=total*2;--p,count--;)
		*p=(*p*1024+irc_calibration/2)/irc_calibration;
	return total*PAIRSIZE;
}

static long irc_write(struct inode *inode, struct file *file, const char *buf,
		     unsigned long count)
{
unsigned long *p=(void *)buf,val;
#ifdef USE_CLI
unsigned long flags;
#endif
long retval=count&~(PAIRSIZE-1);
unsigned char port;
	count/=PAIRSIZE;
	count*=2;
	port=LP_PINITP;
#ifdef USE_CLI
	save_flags(flags);
	cli();
#endif
	while (count--) {
		w_ctr(port);
		if (get_user(val,p++))
			return -EFAULT;
		udelayp(val);
		port^=LP_PAUTOLF;
		}
	w_ctr(LP_PINITP);
#ifdef USE_CLI
	restore_flags(flags);
#endif
	return retval;
}

static int irc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg)
{
	switch (cmd) {

		case IRCON_SET_RTIMEOUT: /* Set the read time-out value. */
			if (!suser())
				return -EINVAL;
			irc_read_timeout=FIXUSEC(arg);
			return 0;

		case IRCON_GET_RTIMEOUT: /* Return the current time-out */
			return put_user(BACKUSEC(irc_read_timeout),(unsigned long *)arg);

		default:
			return -EINVAL;
		}
}
/*
 *	We enforce only one user at a time here with the open/close.
 */

static int irc_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	return 0;
}

static int irc_release(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

/*
 *	The various file operations we support.
 */

static struct file_operations irc_fops = {
	irc_llseek,
	irc_read,
	irc_write,
	NULL,		/* No readdir */
	NULL,		/* No poll */
	irc_ioctl,
	NULL,		/* No mmap */
	irc_open,
	irc_release
};

static struct miscdevice irc_dev =
{
	IRCON_MINOR,
	"ircon",
	&irc_fops
};

#ifdef MODULE
MODULE_PARM(base_io,"i");

void cleanup_module(void)
{
	if (init_level>=ILEV_MISCREG) {
		w_ctr(old_ppctr);
		if (misc_deregister(&irc_dev))
			printk(KERN_ERR "ircon: Huh? misc_deregister failing? Now?\n");
		release_region(IRC_PORT(0),IRC_IO_EXTENT);
		}
}

#define ircon_init init_module
#endif MODULE

__initfunc(int ircon_init(void))
{
unsigned long jiffy;
	if (check_region(IRC_PORT(0), IRC_IO_EXTENT)) {
		printk(KERN_ERR "ircon: Ports already allocated (0x%x-0x%x)\n",
			IRC_PORT(0),IRC_PORT(0)+IRC_IO_EXTENT-1);
		return -EIO;
		}
	if (misc_register(&irc_dev)) {
		printk(KERN_ERR "ircon: misc_register-ing failed!\n");
		return -EIO;
		}
	init_level=ILEV_MISCREG;
	request_region(IRC_PORT(0), IRC_IO_EXTENT, "ircon");
	/* Initialize port and should also check for loopback wire (not yet - hw) */
	old_ppctr=r_ctr();
	w_ctr(LP_PINITP);
	printk(KERN_INFO "Infra-Red Remote Controller Driver v%s - ", IRC_VERSION);
	custom_timeout=1;
	for (jiffy=jiffies;jiffy==jiffies;); /* sync on edge */
	wait_chg((unsigned char)~LP_PBUSY); /* always time-out */
	irc_calibration=wait_chg_ret*1024/(1000000/HZ);
	printk("%lu cycles per sec\n",wait_chg_ret*HZ);
	return 0;
}
