#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#include "util.h"
#include "ircode-db.h"
#include "ircon.h"

#define DEVICE_NAME "/dev/ircon"
#define CHAT_TIMEOUT (15*60)

static char linebuf[LINE_MAX];
static unsigned requests;
static int ircon_fd;

/* in longs! */
#define INIT_PUSHBUF_SIZ 4096
#define ONE2TWO unsigned short userhigh=userdata>>16U,userlow=userdata&0xFFFFU

unsigned long *fbufbase,fbufsiz,fbufpos;

#define sizepushbuf(siz) chk(fbufbase=realloc(fbufbase,(siz)*sizeof(long)))
#define retpush() do { pushbuf(0); fbufpos&=~1; *fbufbase=fbufpos/2; sizepushbuf(fbufpos); return fbufbase; } while (0)

static void pushbuf(unsigned long val)
{
	if (fbufpos>=fbufsiz||!fbufpos)
		sizepushbuf(fbufsiz*=2);
	fbufbase[fbufpos++]=val;
}

static inline void pushinit(void)
{ fbufbase=NULL; fbufsiz=INIT_PUSHBUF_SIZ/2; fbufpos=0; }

static void chkbits(const char *cs)
{
	for (;*cs;cs++)
		if (*cs!='0'&&*cs!='1') {
			fprintf(stderr,"Non-digital characted in input: \"%s\"!\n",cs);
			exit(EXIT_FAILURE);
			}
}
/*** Three-Zero / One-Zero encoder ***/
/* Converts 0->01 and 1->0001 */

static void *filt_tzoz(void *curdata,unsigned userdata)
{ /* I:string, O:string */
char *newdata,*s;
const char *cs;
unsigned tot=0;
#define CURDATA ((const char *)curdata)
	chkbits(CURDATA);
	for (cs=CURDATA;*cs;cs++) tot+=*cs-'0'+1;
	chk(newdata=malloc(tot*2+1));
	for (s=newdata;*CURDATA;CURDATA++) {
		if (*CURDATA=='1') { *s++='0'; *s++='0'; }
		*s++='0'; *s++='1';
		}
	*s='\0';
	return newdata;
#undef CURDATA
}

/*** Header prefixing encoder ***/
/* newdata = userlow times 1 + userhigh times 0 + curdata */

static void *filt_header(void *curdata,unsigned userdata)
{ /* I:string, O:string */
char *newdata,*s;
ONE2TWO;
#define CURDATA ((const char *)curdata)
	chkbits(CURDATA);
	chk(newdata=malloc(strlen(CURDATA)+userlow+userhigh+1));
	for (s=newdata;userlow ;userlow --) *s++='1';
	for (         ;userhigh;userhigh--) *s++='0';
	strcpy(s,CURDATA);
	return newdata;
#undef CURDATA
}

/*** Header prefixing, data duplicating/inverting encoder ***/
/* newdata = dummy "0" + userdata + curdata + "*" + userdata + inverted curdata */
/* Very specific encapsulation for Sharp. */

static void *filt_sharp(void *curdata,unsigned userdata)
{ /* I:string, O:string */
char *newdata,*s;
const char *cs;
#define CURDATA ((const char *)curdata)
#define USERDATA ((const char *)userdata)
	chkbits(CURDATA);
	chk(newdata=malloc((strlen(CURDATA)+strlen(USERDATA))*2+3));
	s=newdata;
	*s++='0';
	for (cs=USERDATA;*cs;cs++) *s++=*cs;
	for (cs= CURDATA;*cs;cs++) *s++=*cs;
	*s++='*';
	for (cs=USERDATA;*cs;cs++) *s++=*cs;
	for (cs= CURDATA;*cs;cs++) *s++=*cs=='0'?'1':'0';
	*s='\0';
	return newdata;
#undef CURDATA
#undef USERDATA
}

/*** Byte duplicating/inverting encoder ***/
/* newdata = userdata times curdata + offseted curdata0..7 + inverted curdata0..7 + curdata8..15 ... */
/* Used in Maspro/Toshiba and AIWA encoding. */

static void *filt_byteinv(void *curdata,unsigned userdata)
{ /* I:string, O:string */
char *newdata,*s;
const char *cs;
#define CURDATA ((const char *)curdata)
	chkbits(CURDATA);
	chk(newdata=malloc(strlen(CURDATA)*2+1));
	for (s=newdata;*CURDATA&&userdata;userdata--)
		*s++=*CURDATA++;
	while (*CURDATA) {
		for (cs=CURDATA;*cs&&cs<CURDATA+8;cs++) *s++=*cs;
		for (;CURDATA<cs;CURDATA++) *s++=*CURDATA=='0'?'1':'0';
		}
	*s='\0';
	return newdata;
#undef CURDATA
}

/*** Delay generating encoder ***/
/* curdata width low pulse, 0 high pulse */

static void *filt_delay(void *curdata,unsigned userdata)
{ /* I:string, O:twinlongs */
int tm;
#define CURDATA ((const char *)curdata)
	tm=atoi(CURDATA);
	sleep(tm/1000000);
	pushinit();
	pushbuf(0); pushbuf(0); /* dummy */
	pushbuf(IRCONV(tm%1000000)); pushbuf(0);
	retpush();
#undef CURDATA
}

/*** String to fixed-width still pulse encoder ***/
/* 0->userlow  width low  pulse */
/* 1->userhigh width high pulse */
/* Used in higher-level (slower) modulation */

static void *filt_slow(void *curdata,unsigned userdata)
{ /* I:string, O:twinlongs */
unsigned long val;
ONE2TWO;
#define CURDATA ((const char *)curdata)
	chkbits(CURDATA);
	pushinit();
	while (*CURDATA) {
		for (val=0;*CURDATA=='0';CURDATA++) val+=userlow ;
		pushbuf(val);
		for (val=0;*CURDATA=='1';CURDATA++) val+=userhigh;
		pushbuf(val);
		}
	retpush();
#undef CURDATA
}

/*** Fixed-pulse, variable spaced encoder ***/
/* 0->low0 width low pulse, high width high pulse */
/* 1->low1 width low pulse, high width high pulse */
/* Used in Korting TV encoding */

struct filt_puls_par {
	unsigned long low0,low1,high,spec;
	};

static void *filt_puls(void *curdata,unsigned userdata)
{ /* I:string, O:twinlongs */
#define CURDATA ((const char *)curdata)
#define USERDATA ((struct filt_puls_par *)userdata)
	pushinit();
	while (*CURDATA) {
		pushbuf(*CURDATA=='0'?USERDATA->low0:(*CURDATA=='*'?USERDATA->spec:USERDATA->low1));
		pushbuf(USERDATA->high);
		CURDATA++;
		}
	retpush();
#undef CURDATA
#undef USERDATA
}

/*** Amplitude-modulating encoder ***/
/* Modulation signal has userlow low usecs and userhigh high usecs. */
/* Used in low-level (faster) modulation */

static void *filt_AM(void *curdata,unsigned userdata)
{ /* I:twinlongs, O:twinlongs */
long cnt,inpsiz;
unsigned long first;
ONE2TWO;
#define CURDATA ((unsigned long *)curdata)
	pushinit();
	inpsiz=*CURDATA; *CURDATA=0;
	while (inpsiz--) {
		first=*CURDATA++;
		cnt=*CURDATA++;
		do {
			pushbuf(userlow+first); first=0;
			pushbuf(userhigh);
			cnt-=userlow+userhigh;
			} while (cnt>0);
		}
	retpush();
#undef CURDATA;
}

/*** Twinlongs signal duplicating encoder ***/
/* Duplicates signal with userdata inserted space. */
/* Needed by Korting TV. */

static void *filt_duptwin(void *curdata,unsigned userdata)
{ /* I:twinlongs, O:twinlongs */
unsigned long *newdata;
size_t siz;
#define CURDATA ((unsigned long *)curdata)
	chk(newdata=malloc((siz=*CURDATA*2*sizeof(*CURDATA))*2));
	memcpy(newdata,CURDATA,siz);
	memcpy(newdata+(*newdata*=2),CURDATA,siz);
	newdata[*newdata]=userdata;
	return newdata;
#undef CURDATA;
}

/*** Raw data transmitter ***/

static void *filt_xmit(void *curdata,unsigned userdata)
{ /* I:twinlongs, O:NULL */
size_t totsiz,written;
#define CURDATA ((unsigned long *)curdata)
	totsiz=*CURDATA*2*sizeof(*CURDATA);
	*CURDATA=0;
	if (!CURDATA[1]&&totsiz) {
		CURDATA+=2;
		totsiz-=2*sizeof(*CURDATA);
		}
	if ((written=write(ircon_fd,CURDATA,totsiz))!=totsiz) {
		fprintf(stderr,"ircond: Write to device \"%s\" truncated - %u of %u written (%s)!\n",DEVICE_NAME,written,totsiz,strerror(errno));
		exit(EXIT_FAILURE);
		}
	return NULL;
#undef CURDATA;
}

struct filt_puls_par /* ShortPulse parameters - see definitions below. WARNING! USHORTs! */
korting_puls_par={IRCONV(4970),IRCONV(7030),IRCONV(168),IRCONV(    0)},     /* No special. */
  sharp_puls_par={IRCONV( 800),IRCONV(1860),IRCONV(260),IRCONV(45000)}; /* Inter-block space. */

static const struct {
	const char *vendor;
	void *(*func)(void *,unsigned);
#define TWO2ONE(a,b) (((b)<<16)|(a))
#define TWODUP(a) TWO2ONE((a),(a))
	unsigned userdata;
	} xmittab[]={
/* Generate usec delay */
{"Delay",filt_delay,0},{NULL,filt_xmit,0},
/* Three-Zero/One-Zero -> Header prefix -> FixPulse 420usec (360..430..470) -> AM 16usec/10usec (14..17/9..12:16/10,15/11) -> Transmit */
{"Panasonic",filt_tzoz,0},{NULL,filt_header,TWO2ONE(8,1)},{NULL,filt_slow,TWODUP(IRCONV(415))},
	{NULL,filt_AM,TWO2ONE(IRCONV(16),IRCONV(10))},{NULL,filt_xmit,0},
/* Dummy "0" header prefix -> ShortPulse 0->4970usec, 1->7030usec, high 168usec -> AM 10usec/18usec -> Duplication w/14230usec space -> Transmit */
{"Korting",filt_header,TWO2ONE(0,1)},{NULL,filt_puls,(unsigned)&korting_puls_par},
	{NULL,filt_AM,TWO2ONE(IRCONV(10),IRCONV(18))},{NULL,filt_duptwin,IRCONV(14230)},{NULL,filt_xmit,0},
/* Last byte inversion/duplicate -> Three-Zero/One-Zero -> Header prefix -> FixPulse 540usec -> AM 18usec/7usec -> Transmit */
{"AIWA",filt_byteinv,27},{NULL,filt_tzoz,0},{NULL,filt_header,TWO2ONE(17,5)},{NULL,filt_slow,TWODUP(IRCONV(540))},
	{NULL,filt_AM,TWO2ONE(IRCONV(19),IRCONV(8))},{NULL,filt_xmit,0},
/* Byte inversion/duplication -> Header prefix -> Three-Zero/One-Zero -> Header prefix -> FixPulse 545usec/530usec -> AM 15usec/10usec -> Transmit */
{"Maspro",filt_byteinv,0},{NULL,filt_header,TWO2ONE(1,0)},{NULL,filt_tzoz,0},{NULL,filt_header,TWO2ONE(16,5)},
/* 490..550/530 */
/* 520/490..560 */
/* 15/10 or 14/11 */
/* 13..14/11 or 14..15/10 */
/* 13/11 relatively good?! */
/* 12.50..14.50/11 -> 13.50/11 */
/* 13.50/10.25..12.25 -> 13.50/11.25 */
/* 13.50/11.25: 505..545 -> 525 */
	{NULL,filt_slow,TWODUP(IRCONV(525))},{NULL,filt_AM,TWO2ONE(IRCONV(13.25),IRCONV(11.25))},{NULL,filt_xmit,0},
/* Byte inversion/duplication -> Header prefix -> Three-Zero/One-Zero -> Header prefix -> FixPulse 545usec/530usec -> AM 19usec/7usec -> Transmit */
{"Toshiba",filt_byteinv,0},{NULL,filt_header,TWO2ONE(1,0)},{NULL,filt_tzoz,0},{NULL,filt_header,TWO2ONE(16,5)},
	{NULL,filt_slow,TWO2ONE(IRCONV(545),IRCONV(530))},{NULL,filt_AM,TWO2ONE(IRCONV(19),IRCONV(7))},{NULL,filt_xmit,0},
/* Dummy "0" header prefix -> ShortPulse 0->4970usec, 1->7030usec, high 168usec -> AM 13usec -> Transmit */
{"Sharp",filt_sharp,(unsigned)"11001"},{NULL,filt_puls,(unsigned)&sharp_puls_par},
	{NULL,filt_AM,TWO2ONE(IRCONV(15),IRCONV(11))},{NULL,filt_xmit,0}
};

#define XMITTABSIZ (sizeof(xmittab)/sizeof(*xmittab))

static void xmit(const char *name)
{
int idx;
const char *cs,*bits;
size_t len;
void *curdata,*newdata;
	if (!strcmp(name,"list")) {
			irdb_dump();
			return;
			}
	if (!(bits=irdb_getbits(name))) {
		printf("Unknown action/command \"%s\" - ignored!\n",name);
		return;
		}
	if (!isdigit(*bits)) {
char *s;
const char *cs2;
		printf("Multi-command %s, expanding {\n",name);
		for (cs=bits;*cs;cs=cs2) {
			for (cs2=cs;*cs2&&*cs2!=',';cs2++);
			chk(s=malloc((len=cs2-cs)+1));
			memcpy(s,cs,len);
			s[len]='\0';
			xmit(s);
			free(s);
			if (*cs2) cs2++;
			}
		printf("}\n");
		return;
		}
	for (cs=name;*cs&&*cs!='-';cs++);
	if (!cs) {
		printf("Vendor ID not separable from \"%s\" - ignored!\n",name);
		return;
		}
	len=cs-name;
	for (idx=0;idx<XMITTABSIZ;idx++)
		if (xmittab[idx].vendor&&!strncmp(name,xmittab[idx].vendor,len)&&!xmittab[idx].vendor[len])
			break;
	if (idx==XMITTABSIZ) {
		fprintf(stderr,"ircond: Handler for vendor of \"%s\" unavailable, update me!\n",name);
		printf("Action \"%s\" unavailable!\n",name);
		return;
		}
	printf("Transmitting %s.",name); fflush(stdout);
	if ((ircon_fd=open(DEVICE_NAME,O_WRONLY))==-1) {
		fprintf(stderr,"ircond: opendevice(\"%s\") for write failed (%s)!\n",DEVICE_NAME,strerror(errno));
		exit(EXIT_FAILURE);
		}
	curdata=strdup(bits);
	putchar('.'); fflush(stdout);
	do {
		newdata=(*xmittab[idx].func)(curdata,xmittab[idx].userdata);
		free(curdata); curdata=newdata;
		idx++;
		} while (idx<XMITTABSIZ&&!xmittab[idx].vendor);
	putchar('.'); fflush(stdout);
	free(curdata);
	if (close(ircon_fd)) {
		fprintf(stderr,"ircond: closedevice(\"%s\") failed (%s)!\n",DEVICE_NAME,strerror(errno));
		exit(EXIT_FAILURE);
		}
	printf("done\n");
	requests++;
}

int main(int argc,char **argv)
{
char peername[32]="unidentified/local host";
struct sockaddr_in saddr;
int saddrlen=sizeof(saddr),dohelp=1;
char *s;

	irdb_load(NULL);
	if (--argc) {
		while (argc--)
			xmit(*++argv);
		return(EXIT_SUCCESS);
		}
	if (!getpeername(STDIN_FILENO,(struct sockaddr *)&saddr,&saddrlen)&&saddr.sin_family==AF_INET)
		sprintf(peername,"%08lx:%u",ntohl(saddr.sin_addr.s_addr),saddr.sin_port);
	for (;;) {
		alarm(CHAT_TIMEOUT);
		if (dohelp) {
			printf("\
Hi %s, pleased to see you again!\n\
We're glad to offer you remote infra-red control services as we are -\n\
 - The Infra-Red Remote Controller Daemon! (version 0.99)\n\
\n\
Available commands: help,list,quit==exit\n\
",peername);
			dohelp=0;
			}

		putchar('>'); fflush(stdout);
		if (!fgets(linebuf,sizeof(linebuf),stdin)) {
out:
			printf("\nBailing out after %u served requests. See ya!\n",requests);
			return(EXIT_SUCCESS);
			}
		alarm(0);
		if (!(s=strchr(linebuf,'\n'))) {
			printf("\nInput line not terminated!\n");
			return(EXIT_FAILURE);
			}
		*s='\0';
		if ((s=strchr(linebuf,'\r'))) *s='\0';
		if (!*linebuf) continue;
		if (!strcmp(linebuf,"quit")||!strcmp(linebuf,"exit"))
			goto out;
		if (!strcmp(linebuf,"help")) {
			dohelp=1;
			continue;
			}
		xmit(linebuf);
		}
	return(EXIT_SUCCESS);
}
