#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "util.h"

#define NUM_CH 80
#define FREQ_THR 11850

static char buf[LINE_MAX];
static const char sats[]="ACEFHJPU";

struct channel {
	short freq;
	enum { AB_A, AB_B } ab;
	enum { DEV_16, DEV_25 } dev;
	char skew[3],sat[3];
	enum { AU_NORM, AU_WIDE, AU_STER } au;
	unsigned left,right;
	};

static int neqchan(struct channel *a,struct channel *b)
{
	return (a->freq!=b->freq||a->ab!=b->ab||a->dev!=b->dev
	  ||a->au!=b->au||a->left!=b->left||(a->au==AU_STER&&a->right!=b->right)
	  ||strcmp(a->skew,b->skew)||strcmp(a->sat,b->sat));
}

static void outcmd(const char *cmd)
{
	printf("Maspro-SRE-100R-%s\nMaspro-SRE-100R-BIGDELAY\n",cmd);
}

static void outnum(unsigned num,int width)
{
char cmd[]="Digit-X",buf[10],fmt[10],*s;
	sprintf(fmt,"%%0%uu",width);
	sprintf(buf,fmt,num);
	for (s=buf;*s;s++) {
		cmd[6]=*s;
		outcmd(cmd);
		}
}

static unsigned parseau(char *au)
{
int cons;
unsigned res;
	if (strlen(au)!=4||au[1]!='.') errout("Audio info \"%s\" must be in format X.xx",au);
	au[1]=au[0]; au++;
	if (sscanf(au,"%u%n",&res,&cons)!=1||cons!=3) errout("Error parsing audio info: \"%s\"",au);
	if (res<600||res>850) errout("Audio info \"%s\" not in range (6.00..8.50)",au);
	return res;
}

static int getchan(const char *fn,struct channel **chp)
{
FILE *f;
int cnt=0,cons;
struct channel *ch;
unsigned freq,dev;
char bufA[3],bufAMod[6],bufLR[11],bufSk[4],bufSa[4];
	if (!(f=fopen(fn,"rt"))) errout("Error opening \"%s\"",fn);
	chk(ch=malloc(sizeof(*ch)*NUM_CH));
	*chp=ch;
	while (fgets(buf,sizeof(buf),f)) {
		procbuf(buf);
		if (!*buf) continue;
		if (sscanf(buf,"%u%2s%u%5s%10s%3s%3s%n",&freq,bufA,&dev,bufAMod,bufLR,bufSk,bufSa,&cons)!=7||buf[cons])
			errout("Error parsing channel \"%s\" in file \"%s\"",buf,fn);
		if (freq<10950||(freq>11750&&freq<11950)||freq>12950) errout("Frequency not in range: %u",freq);
		ch->freq=freq;
		if (strcmp(bufA,"A")&&strcmp(bufA,"B")) errout("Input type \"%s\" must be A or B",bufA);
		ch->ab=bufA[0]=='A'?AB_A:AB_B;
		if (dev!=16&&dev!=25) errout("Deviation %u must be 16 or 25 (MHz)",dev);
		ch->dev=dev==16?DEV_16:DEV_25;
		if (!strcmp(bufAMod,"STER")) {
			if (strlen(bufLR)!=9||bufLR[4]!='/') errout("Stereo information \"%s\" must be in format \"X.xx/Y.yy\"",bufLR);
			bufLR[4]='\0';
			ch->right=parseau(bufLR+5);
			ch->au=AU_STER;
			}
		else {
			ch->au=!strcmp(bufAMod,"WIDE")?AU_WIDE:AU_NORM;
			if (ch->au!=AU_WIDE&&strcmp(bufAMod,"NORM")) errout("Audio-type \"%s\" must be NORM, WIDE or STER",bufAMod);
			}
		ch->left=parseau(bufLR);
		if (strlen(bufSk)!=2||(bufSk[0]!='H'&&bufSk[0]!='V')||bufSk[1]<'1'||bufSk[1]>'4') errout("Skew \"%s\" must be [HV][1..4]",bufSk);
		strcpy(ch->skew,bufSk);
		if (strlen(bufSa)!=2||!strchr(sats,bufSa[0])||bufSk[1]<'1'||bufSk[1]>'3') errout("Satellite \"%s\" must be [%s][1..3]",bufSa,sats);
		strcpy(ch->sat,bufSa);
		cnt++; ch++;
		}
	fclose(f);
	return cnt;
}

int main(int argc,char **argv)
{
struct channel *o,*n;
int no,nn,curch,active=0,chg;
	pname=argv[0];
	if (argc!=3) {
		fprintf(stderr,"Usage: masprog <origdata> <newdata>\n");
		exit(EXIT_FAILURE);
		}
	no=getchan(argv[1],&o);
	nn=getchan(argv[2],&n);
	if (nn>no) errout("Cannot get original channel parameters for %d-%d",no+1,nn+1);
	for (curch=1;nn--;o++,n++,curch++) {
		if (!neqchan(o,n)) continue;
		if (!active) {
			outcmd("Power"); outcmd("Power"); /* outcmd("Test-Mode"); - really not a good idea */
			active=1;
			}
		outnum(curch,2);
		outcmd("Mode");
		chg=0;
		if ((o->freq<FREQ_THR)!=(n->freq<FREQ_THR)) chg=1,outcmd("Recall");
		o->freq-=o->freq<FREQ_THR?10000:11000;
		n->freq-=n->freq<FREQ_THR?10000:11000;
		if (o->freq!=n->freq) chg=1,outnum(n->freq,0);
		if (o->ab!=n->ab) chg=1,outcmd("Parental-Lock");
		if (o->dev!=n->dev) chg=1,outcmd("Wide-Normal");
		while (strcmp(o->skew,n->skew)) {
			chg=1,outcmd("Vertical-Horizontal");
			if (n->skew[1]<'4') n->skew[1]++;
			else { n->skew[0]=n->skew[0]=='H'?'V':'H'; n->skew[1]='1'; }
			}
		while (strcmp(o->sat,n->sat)) {
			chg=1,outcmd("Skew-Up");
			if (n->sat[1]<'3') n->sat[1]++;
			else {
				if (!(n->sat[0]=*(strchr(sats,n->sat[0])+1))) n->sat[0]=sats[0];
				n->sat[1]='1';
				}
			}
		if (chg) { outcmd("Store"); printf("Delay-Maspro-SRE-100R-Store\n"); }
		outcmd("Mode");
		chg=0;
		if (o->au!=n->au) {
			chg=1;
			if (n->au==AU_STER) { outcmd("Channel-Down"); outcmd("Mode"); }
			else {
				if (o->au==AU_STER) outcmd("Channel-Up");
				outcmd("Mode");
				if (o->au!=AU_STER) outcmd("Wide-Normal");
				}
			}
		else outcmd("Mode");
		if (n->au==AU_STER) {
			if (o->au!=AU_STER||o->right!=n->right) chg=1,outnum(n->right,3);
			outcmd("Mode");
			}
		if (o->left!=n->left) chg=1,outnum(n->left,3);
		if (chg) { outcmd("Store"); printf("Delay-Maspro-SRE-100R-Store\n"); }
		outcmd("Mode");
		}
	if (active) { outnum(1,2); outcmd("Power"); outcmd("Power"); }
	return EXIT_SUCCESS;
}
